@@ -13,21 +13,18 @@
 ##############################################################################
 """Command-line option parsing
 """
-from __future__ import print_function
 
 import argparse
-import re
 import os
+import re
 import sys
 
 import pkg_resources
 
-from zope.testrunner.formatter import (
-    ColorfulOutputFormatter,
-    OutputFormatter,
-    SubunitOutputFormatter,
-    SubunitV2OutputFormatter,
-)
+from zope.testrunner.formatter import ColorfulOutputFormatter
+from zope.testrunner.formatter import OutputFormatter
+from zope.testrunner.formatter import SubunitOutputFormatter
+from zope.testrunner.formatter import SubunitV2OutputFormatter
 from zope.testrunner.formatter import terminal_has_colors
 from zope.testrunner.profiling import available_profilers
 
@@ -135,7 +132,7 @@ searching.add_argument(
     default=1,
     help="""\
 Run the tests at the given level.  Any test at a level at or below
-this is run, any test at a level above this is not run.  Level 0
+this is run, any test at a level above this is not run.  Level <= 0
 runs all tests.
 """)
 
