@@ -13,12 +13,13 @@
 ##############################################################################
 """Test execution
 """
-from __future__ import print_function
 
 import errno
 import gc
 import io
 import os
+import pprint
+import queue
 import re
 import subprocess
 import sys
@@ -27,38 +28,38 @@ import time
 import traceback
 import unittest
 import warnings
-
 from contextlib import contextmanager
+from io import StringIO
 
-from six import StringIO
-from zope.testrunner.find import import_name
-from zope.testrunner.find import name_from_layer, _layer_name_cache
-from zope.testrunner.layer import UnitTests
-from zope.testrunner.layer import EmptyLayer
-from zope.testrunner.layer import EmptySuite
-from zope.testrunner.options import get_options
-from zope.testrunner.refcount import TrackRefs
 import zope.testrunner
-import zope.testrunner.coverage
 import zope.testrunner._doctest
-import zope.testrunner.logsupport
-import zope.testrunner.selftest
-import zope.testrunner.profiling
+import zope.testrunner.coverage
+import zope.testrunner.debug
 import zope.testrunner.filter
 import zope.testrunner.garbagecollection
-import zope.testrunner.listing
-import zope.testrunner.statistics
-import zope.testrunner.process
 import zope.testrunner.interfaces
-import zope.testrunner.debug
-import zope.testrunner.tb_format
+import zope.testrunner.listing
+import zope.testrunner.logsupport
+import zope.testrunner.process
+import zope.testrunner.profiling
+import zope.testrunner.selftest
 import zope.testrunner.shuffle
+import zope.testrunner.statistics
+import zope.testrunner.tb_format
+from zope.testrunner import threadsupport
+from zope.testrunner.find import _layer_name_cache
+from zope.testrunner.find import import_name
+from zope.testrunner.find import name_from_layer
+from zope.testrunner.layer import EmptyLayer
+from zope.testrunner.layer import EmptySuite
+from zope.testrunner.layer import UnitTests
+from zope.testrunner.options import get_options
+from zope.testrunner.refcount import TrackRefs
+
+from .digraph import DiGraph
+from .util import is_jython
+from .util import uses_refcounts
 
-try:
-    import Queue # Python 2
-except ImportError:
-    # Python 3
-    import queue as Queue # Python 3
 
 class UnexpectedSuccess(Exception):
     pass
@@ -66,8 +67,6 @@ class UnexpectedSuccess(Exception):
 
 PYREFCOUNT_PATTERN = re.compile(r'\[[0-9]+ refs\]')
 
-is_jython = sys.platform.startswith('java')
-
 
 class SubprocessError(Exception):
     """An error occurred when running a subprocess
@@ -78,14 +77,14 @@ class SubprocessError(Exception):
         self.stderr = stderr
 
     def __str__(self):
-        return '%s: %s' % (self.reason, self.stderr)
+        return '{}: {}'.format(self.reason, self.stderr)
 
 
 class CanNotTearDown(Exception):
     "Couldn't tear down a test"
 
 
-class Runner(object):
+class Runner:
     """The test runner.
 
     It is the central point of this package and responsible for finding and
@@ -94,11 +93,12 @@ class Runner(object):
 
     .. versionchanged:: 4.8.0
        Add the *warnings* keyword argument. If this is ``None`` (the default)
-       and the user hasn't configured Python otherwise with command-line arguments
-       or environment variables, we will enable the default warnings, including
-       ``DeprecationWarning``, when running tests. Otherwise, it can be any
-       string acceptable to :func:`warnings.simplefilter` and that filter will
-       be in effect while running tests.
+       and the user hasn't configured Python otherwise with command-line
+       arguments or environment variables, we will enable the default
+       warnings, including ``DeprecationWarning``, when running tests.
+       Otherwise, it can be any string acceptable to
+       :func:`warnings.simplefilter` and that filter will be in effect while
+       running tests.
 
     """
 
@@ -148,8 +148,8 @@ class Runner(object):
             # but only if this is not in the subprocess
             yield (name_from_layer(EmptyLayer), EmptyLayer, EmptySuite())
 
-        layer_names = dict([(layer_from_name(layer_name), layer_name)
-                            for layer_name in self.tests_by_layer_name])
+        layer_names = {layer_from_name(layer_name): layer_name
+                       for layer_name in self.tests_by_layer_name}
         for layer in order_by_bases(layer_names):
             layer_name = layer_names[layer]
             yield layer_name, layer, self.tests_by_layer_name[layer_name]
@@ -180,9 +180,9 @@ class Runner(object):
 
             # Late setup
             #
-            # Some system tools like profilers are really bad with stack frames.
-            # E.g. hotshot doesn't like it when we leave the stack frame that we
-            # called start() from.
+            # Some system tools like profilers are really bad with stack
+            # frames.  E.g. hotshot doesn't like it when we leave the stack
+            # frame that we called start() from.
             for feature in self.features:
                 feature.late_setup()
 
@@ -268,9 +268,10 @@ class Runner(object):
                 # noisy.  The -Wd and -Wa flags can be used to bypass this
                 # only when self.warnings is None.
                 if self.warnings in ['default', 'always']:
-                    warnings.filterwarnings('module',
-                                            category=DeprecationWarning,
-                                            message=r'Please use assert\w+ instead.')
+                    warnings.filterwarnings(
+                        'module',
+                        category=DeprecationWarning,
+                        message=r'Please use assert\w+ instead.')
             yield
 
     def run_tests(self):
@@ -375,7 +376,9 @@ def run_tests(options, tests, name, failures, errors, skipped, import_errors):
                         test.debug()
                     except KeyboardInterrupt:
                         raise
-                    except:
+                    except unittest.SkipTest as e:
+                        result.addSkip(test, str(e))
+                    except BaseException:
                         result.addError(
                             test,
                             sys.exc_info()[:2] + (sys.exc_info()[2].tb_next, ),
@@ -401,11 +404,8 @@ def run_tests(options, tests, name, failures, errors, skipped, import_errors):
         output.stop_tests()
         failures.extend(result.failures)
         n_failures = len(result.failures)
-        if hasattr(result, 'unexpectedSuccesses'):
-            # Python versions prior to 2.7 do not have the concept of
-            # unexpectedSuccesses.
-            failures.extend(result.unexpectedSuccesses)
-            n_failures += len(result.unexpectedSuccesses)
+        failures.extend(result.unexpectedSuccesses)
+        n_failures += len(result.unexpectedSuccesses)
         skipped.extend(result.skipped)
         errors.extend(result.errors)
         output.summary(n_tests=result.testsRun,
@@ -452,7 +452,7 @@ def run_layer(options, layer_name, layer, tests, setup_layers,
     output = options.output
     gathered = []
     gather_layers(layer, gathered)
-    needed = dict([(l, 1) for l in gathered])
+    needed = {ly: 1 for ly in gathered}
     if options.resume_number != 0:
         output.info("Running %s tests:" % layer_name)
     tear_down_unneeded(options, needed, setup_layers, errors)
@@ -479,7 +479,7 @@ class SetUpLayerFailure(unittest.TestCase):
     subunit_label = 'setUp'
 
     def __init__(self, layer):
-        super(SetUpLayerFailure, self).__init__()
+        super().__init__()
         self.layer = layer
 
     def runTest(self):
@@ -494,7 +494,7 @@ class TearDownLayerFailure(unittest.TestCase):
     subunit_label = 'tearDown'
 
     def __init__(self, layer):
-        super(TearDownLayerFailure, self).__init__()
+        super().__init__()
         self.layer = layer
 
     def runTest(self):
@@ -521,11 +521,6 @@ def spawn_layer_in_subprocess(result, script_parts, options, features,
 
         args.extend(options.original_testrunner_args[1:])
 
-        # this is because of a bug in Python (http://www.python.org/sf/900092)
-        if (options.profile == 'hotshot'
-            and sys.version_info[:3] <= (2, 4, 1)):
-            args.insert(1, '-O')
-
         debugargs = args  # save them before messing up for windows
         if sys.platform.startswith('win'):
             args = args[0] + ' ' + ' '.join([
@@ -535,10 +530,22 @@ def spawn_layer_in_subprocess(result, script_parts, options, features,
         for feature in features:
             feature.layer_setup(layer)
 
-        child = subprocess.Popen(args, shell=False, stdin=subprocess.PIPE,
+        child = subprocess.Popen(
+            args, shell=False, stdin=subprocess.PIPE,
             stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd,
             close_fds=not sys.platform.startswith('win'))
 
+        def reader_thread(f, buf):
+            buf.append(f.read())
+
+        # Start reading stderr in a thread.  This means we don't hang if the
+        # subprocess writes more to stderr than the pipe capacity.
+        stderr_buf = []
+        stderr_thread = threading.Thread(
+            target=reader_thread, args=(child.stderr, stderr_buf))
+        stderr_thread.daemon = True
+        stderr_thread.start()
+
         while True:
             try:
                 while True:
@@ -548,11 +555,11 @@ def spawn_layer_in_subprocess(result, script_parts, options, features,
                     # return its lines as a batch). We don't want too much
                     # buffering because this foils automatic and human monitors
                     # trying to verify that the subprocess is still alive.
-                    l = child.stdout.readline()
-                    if not l:
+                    line = child.stdout.readline()
+                    if not line:
                         break
-                    result.write(l)
-            except IOError as e:
+                    result.write(line)
+            except OSError as e:
                 if e.errno == errno.EINTR:
                     # If the subprocess dies before we finish reading its
                     # output, a SIGCHLD signal can interrupt the reading.
@@ -564,8 +571,9 @@ def spawn_layer_in_subprocess(result, script_parts, options, features,
             else:
                 break
 
-        # Now stderr should be ready to read the whole thing.
-        errlines = child.stderr.read().splitlines()
+        # Now we should be able to finish reading stderr.
+        stderr_thread.join()
+        errlines = stderr_buf[0].splitlines()
         erriter = iter(errlines)
         nfail = nerr = 0
         for line in erriter:
@@ -581,7 +589,7 @@ def spawn_layer_in_subprocess(result, script_parts, options, features,
             if options.verbose >= 1:
                 errmsg += "\nChild command line: %s" % debugargs
             if (options.verbose >= 2 or
-                (options.verbose == 1 and len(errlines) < 20)):
+                    (options.verbose == 1 and len(errlines) < 20)):
                 errmsg += ("\nChild stderr was:\n" +
                            "\n".join("  " + line.decode('utf-8', 'replace')
                                      for line in errlines))
@@ -594,7 +602,6 @@ def spawn_layer_in_subprocess(result, script_parts, options, features,
                                      for line in errlines[-10:]))
             output.error_with_banner(errmsg)
 
-
         while nfail > 0:
             nfail -= 1
             # Doing erriter.next().strip() confuses the 2to3 fixer, so
@@ -641,7 +648,7 @@ def _get_output_buffer(stream):
     return stream
 
 
-class AbstractSubprocessResult(object):
+class AbstractSubprocessResult:
     """A result of a subprocess layer run."""
 
     num_ran = 0
@@ -668,7 +675,7 @@ class ImmediateSubprocessResult(AbstractSubprocessResult):
     """Sends complete output to queue."""
 
     def __init__(self, layer_name, queue):
-        super(ImmediateSubprocessResult, self).__init__(layer_name, queue)
+        super().__init__(layer_name, queue)
         self.stream = _get_output_buffer(sys.stdout)
 
     def write(self, out):
@@ -677,7 +684,9 @@ class ImmediateSubprocessResult(AbstractSubprocessResult):
         self.stream.flush()
 
 
-_is_dots = re.compile(br'\.+(\r\n?|\n)').match # Windows sneaks in a \r\n.
+_is_dots = re.compile(br'\.+(\r\n?|\n)').match  # Windows sneaks in a \r\n.
+
+
 class KeepaliveSubprocessResult(AbstractSubprocessResult):
     "Keeps stdout for later processing; sends marks to queue to show activity."
 
@@ -705,7 +714,7 @@ def resume_tests(script_parts, options, features, layers, failures, errors,
     elif (options.verbose > 1 and
             not options.subunit and not options.subunit_v2):
         result_factory = KeepaliveSubprocessResult
-        stdout_queue = Queue.Queue()
+        stdout_queue = queue.Queue()
     else:
         result_factory = DeferredSubprocessResult
     resume_number = int(options.processes > 1)
@@ -742,7 +751,7 @@ def resume_tests(script_parts, options, features, layers, failures, errors,
             previous_output = output
             try:
                 layer_name, output = stdout_queue.get(False)
-            except Queue.Empty:
+            except queue.Empty:
                 break
             if layer_name != last_layer_intermediate_output:
                 # Clarify what layer is reporting activity.
@@ -770,7 +779,7 @@ def resume_tests(script_parts, options, features, layers, failures, errors,
 
         # Help keep-alive monitors (human or automated) keep up-to-date.
         stdout.flush()
-        time.sleep(0.01) # Keep the loop from being too tight.
+        time.sleep(0.01)  # Keep the loop from being too tight.
 
     # Return the total number of tests run.
     return sum(r.num_ran for r in results)
@@ -779,29 +788,30 @@ def resume_tests(script_parts, options, features, layers, failures, errors,
 def tear_down_unneeded(options, needed, setup_layers, errors, optional=False):
     # Tear down any layers not needed for these tests. The unneeded layers
     # might interfere.
-    unneeded = [l for l in setup_layers if l not in needed]
+    unneeded = [layer for layer in setup_layers if layer not in needed]
     unneeded = order_by_bases(unneeded)
     unneeded.reverse()
     output = options.output
-    for l in unneeded:
-        output.start_tear_down(name_from_layer(l))
+    for layer in unneeded:
+        output.start_tear_down(name_from_layer(layer))
         t = time.time()
         try:
             try:
-                if hasattr(l, 'tearDown'):
-                    l.tearDown()
+                if hasattr(layer, 'tearDown'):
+                    layer.tearDown()
             except NotImplementedError:
                 output.tear_down_not_supported()
                 if not optional:
-                    raise CanNotTearDown(l)
+                    raise CanNotTearDown(layer)
             except MemoryError:
                 raise
             except Exception:
-                handle_layer_failure(TearDownLayerFailure(l), output, errors)
+                handle_layer_failure(
+                    TearDownLayerFailure(layer), output, errors)
             else:
                 output.stop_tear_down(time.time() - t)
         finally:
-            del setup_layers[l]
+            del setup_layers[layer]
 
 
 cant_pm_in_subprocess_message = """
@@ -880,13 +890,27 @@ class TestResult(unittest.TestResult):
             if hasattr(layer, 'testTearDown'):
                 layer.testTearDown()
 
+    def _makeBufferedStdStream(self):
+        """Make a buffered stream to replace a standard stream."""
+        # The returned stream needs to have a 'buffer' attribute, since
+        # some tests may expect that to exist on standard streams, and a
+        # 'getvalue' method for the convenience of _restoreStdStreams.
+        # This requires some care.
+        class BufferedStandardStream(io.TextIOWrapper):
+            def getvalue(self):
+                return self.buffer.getvalue().decode(
+                    encoding=self.encoding, errors=self.errors)
+
+        return BufferedStandardStream(
+            io.BytesIO(), newline='\n', write_through=True)
+
     def _setUpStdStreams(self):
         """Set up buffered standard streams, if requested."""
         if self.options.buffer:
             if self._stdout_buffer is None:
-                self._stdout_buffer = StringIO()
+                self._stdout_buffer = self._makeBufferedStdStream()
             if self._stderr_buffer is None:
-                self._stderr_buffer = StringIO()
+                self._stderr_buffer = self._makeBufferedStdStream()
             sys.stdout = self._stdout_buffer
             sys.stderr = self._stderr_buffer
 
@@ -906,15 +930,16 @@ class TestResult(unittest.TestResult):
             return None, None
 
     def startTest(self, test):
+        self._test_state = test.__dict__.copy()
         self.testSetUp()
         unittest.TestResult.startTest(self, test)
-        testsRun = self.testsRun - 1 # subtract the one the base class added
+        testsRun = self.testsRun - 1  # subtract the one the base class added
         count = test.countTestCases()
         self.testsRun = testsRun + count
 
         self.options.output.start_test(test, self.testsRun, self.count)
 
-        self._threads = threading.enumerate()
+        self._threads = threadsupport.enumerate()
         self._start_time = time.time()
 
         self._setUpStdStreams()
@@ -985,13 +1010,38 @@ class TestResult(unittest.TestResult):
                                                       " as a subprocess!")
             else:
                 # XXX: what exc_info? there's no exc_info!
-                zope.testrunner.debug.post_mortem(exc_info)
+                # flake8 is correct, but keep it quiet for now ...
+                zope.testrunner.debug.post_mortem(exc_info)  # noqa: F821
         elif self.options.stop_on_error:
             self.stop()
 
     def stopTest(self, test):
         self.testTearDown()
-        self.options.output.stop_test(test)
+        # Without clearing, cyclic garbage referenced by the test
+        # would be reported in the following test.
+        test.__dict__.clear()
+        test.__dict__.update(self._test_state)
+        del self._test_state
+        cycles = None
+        if (uses_refcounts and self.options.gc_after_test and
+                self.options.verbose >= 4):
+            gc_opts = gc.get_debug()
+            gc.set_debug(gc.DEBUG_SAVEALL)
+            gc.collect()
+            if gc.garbage:
+                g = DiGraph(gc.garbage)
+                for obj in gc.garbage:
+                    g.add_neighbors(obj, gc.get_referents(obj))
+                cycles = [[repr_lines(o) for o in c] for c in g.sccs()]
+                del gc.garbage[:]
+                g = obj = None  # avoid to hold cyclic garbage
+            gc.set_debug(gc_opts)
+        gccount = gc.collect() \
+            if uses_refcounts and self.options.gc_after_test else 0
+        self.options.output.stop_test(test, gccount)
+
+        if cycles:
+            self.options.output.test_cycles(test, cycles)
 
         if is_jython:
             pass
@@ -1003,7 +1053,7 @@ class TestResult(unittest.TestResult):
 
         # Did the test leave any new threads behind?
         new_threads = []
-        for t in threading.enumerate():
+        for t in threadsupport.enumerate():
             if t.is_alive() and t not in self._threads:
                 if not any([re.match(p, t.name)
                             for p in self.options.ignore_new_threads]):
@@ -1043,11 +1093,13 @@ def layer_sort_key(layer):
     Based on the reverse MRO ordering in order to put layers with shared base
     layers next to each other.
     """
-    seen = set([])
+    seen = set()
     key = []
 
     # Note: we cannot reuse gather_layers() here because it uses a
     # different traversal order.
+    binding = {}  # hack to avoid recursion -- for PY2
+
     def _gather(layer):
         seen.add(layer)
         # We make the simplifying assumption that the order of initialization
@@ -1057,11 +1109,15 @@ def layer_sort_key(layer):
         # zope.testrunner, so let's use that.
         for base in layer.__bases__[::-1]:
             if base is not object and base not in seen:
-                _gather(base)
+                binding["_gather"](base)
         key.append(layer)
 
+    binding["_gather"] = _gather
     _gather(layer)
-    return tuple(name_from_layer(l) for l in key if l != UnitTests)
+    try:
+        return tuple(name_from_layer(ly) for ly in key if ly != UnitTests)
+    finally:
+        binding.clear()  # break reference cycle
 
 
 def order_by_bases(layers):
@@ -1097,11 +1153,31 @@ class FakeInputContinueGenerator:
     def readline(self):
         print('c\n')
         print('*'*70)
-        print ("Can't use pdb.set_trace when running a layer"
-               " as a subprocess!")
+        print("Can't use pdb.set_trace when running a layer"
+              " as a subprocess!")
         print('*'*70)
         print()
         return 'c\n'
 
     def close(self):
         pass
+
+
+def repr_lines(obj, max_width=75, max_lines=5):
+    """represent *obj* by a sequence of text lines.
+
+    Use at most *max_lines*, each with at most *max_width* chars.
+    """
+    try:
+        oi = pprint.pformat(obj)
+    except Exception:
+        # unprintable
+        oi = "{} instance at 0x{:x}".format(obj.__class__, id(obj))
+    # limit
+    cmps = oi.split("\n", max_lines)
+    if len(cmps) > max_lines:
+        cmps[-1] = "..."
+    for i, li in enumerate(cmps):
+        if len(li) > max_width:
+            cmps[i] = li[:max_width - 3] + "..."
+    return cmps
